/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "libdspam_objects.h"
#include "libdspam.h"
#include "nodetree.h"
#include "config.h"
#include "base64.h"
#include "util.h"
#include "storage_driver.h"
#include "buffer.h"
#include "lht.h"
#include "tbt.h"
#include "error.h"
#include "decode.h"
#include "language.h"

/* dspam_init: initializes a DSPAM context 
   parameters: const char *dictionary	filename of dictionary
               int mode			processing mode (libdspam.h)
               int flags                processing flags (libdspam.h)

      returns: pointer to the newly allocated DSPAM context 
*/

DSPAM_CTX *
dspam_init (const char *username, const char *group, int mode, int flags)
{
  DSPAM_CTX *CTX;

  CTX = calloc (1, sizeof (DSPAM_CTX));
  if (CTX == NULL)
    return NULL;

  if (username != NULL && username[0] != 0)
    CTX->username = strdup (username);
  else
    CTX->username = NULL;

  if (group != NULL && group[0] != 0)
    CTX->group = strdup (group);
  else
    CTX->group = NULL;
  CTX->probability = -1;
  CTX->mode = mode;
  CTX->flags = flags;
  CTX->message = NULL;
  CTX->confidence = 0;

  if (!_ds_init_storage (CTX))
    return CTX;

  free(CTX->username);
  free(CTX->group);
  free(CTX);
  return NULL;
}


/* dspam_destroy: destroys a DSPAM context
      parameters: DSPAM_CTX *	pointer to context
         returns: 0 on successful completion
*/

int
dspam_destroy (DSPAM_CTX * CTX)
{
  if (CTX->storage != NULL) {

    /* Sanity-Check Totals */
    if (CTX->totals.total_spam < 0)
      CTX->totals.total_spam = 0;
    if (CTX->totals.total_innocent < 0)
      CTX->totals.total_innocent = 0;
    if (CTX->totals.spam_misclassified < 0)
      CTX->totals.spam_misclassified = 0;
    if (CTX->totals.innocent_misclassified < 0)
      CTX->totals.innocent_misclassified = 0;

    _ds_shutdown_storage (CTX);
 
    free (CTX->username);
    free (CTX->group);

    /* Only destroy signature if we created it */
    if (CTX->signature != NULL && CTX->mode == DSM_PROCESS)
    {
      if (CTX->signature->data)
        free (CTX->signature->data);
      free (CTX->signature);
    }
  }

  free (CTX);
  return 0;
}

/* dspam_getresult: returns results of dspam_process() function
        parameters: DSPAM_CTX *	pointer to context
           returns: DSR_ISSPAM     - Is spam
                    DSR_ISINNOCENT - Is not spam 
                    <Other>        - Error code if dspam_process returned a
                                     non-zero value
*/

int
dspam_getresult (DSPAM_CTX * CTX)
{
  return CTX->result;
}

/* dspam_process: processes a message using the specified mode and flags 
      parameters: DSPAM_CTX *			pointer to context
                  const char *message		pointer to message text
         returns: 0 on successful completion
*/
int
dspam_process (DSPAM_CTX * CTX, const char *message)
{
  struct nt_node *node_header, *node_nt;
  struct nt_c c_nt, c_nt2;
  buffer *header, *body;
  char *decode, *x, *y;
  char heading[1024];
  int i = 0, spam_result = 0;
  int is_signed = 0;

  CTX->_process_start = time (NULL);

  if (CTX->mode != DSM_PROCESS && CTX->mode != DSM_CLASSIFY
      && (CTX->flags & DSF_SIGNATURE))
    return _ds_process_signature (CTX);

  header = buffer_create (NULL);
  body = buffer_create (NULL);

  if (header == NULL || body == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    buffer_destroy (header);
    buffer_destroy (body);
    return EUNKNOWN;
  }

  if (CTX->message == NULL && message != NULL)
    CTX->message = _ds_actualize_message (message);

  /* Skip this section is signature was provided for classification */
  if (!(CTX->flags & DSF_SIGNATURE && CTX->mode == DSM_CLASSIFY && CTX->signature != NULL))
  {

    if (CTX->message == NULL)
    {
      LOG (LOG_WARNING, "_ds_actualize_message() failed");
      return EUNKNOWN;
    }

    /* Iterate through each component and create large header/body buffers */

    node_nt = c_nt_first (CTX->message->components, &c_nt);
    while (node_nt != NULL)
    {
      struct _ds_message_block *block =
        (struct _ds_message_block *) node_nt->ptr;

#ifdef VERBOSE
      LOGDEBUG ("Processing component %d", i);
#endif

      if (block->media_type == MT_MULTIPART
          && block->media_subtype == MST_SIGNED)
        is_signed = 1;

      if (block->headers == NULL || block->headers->items == 0)
      {
#ifdef VERBOSE
        LOGDEBUG ("  : End of Message Identifier");
#endif

        /* Skip Attachments */
      }
      else if (block->media_type == MT_TEXT
               || block->media_type == MT_MESSAGE 
               || block->media_type == MT_UNKNOWN || 
                  (i == 0 && (block->media_type == MT_TEXT || 
                              block->media_type == MT_MULTIPART ||
                              block->media_type == MT_MESSAGE) ))
      {

        /* Accumulate the headers */
        node_header = c_nt_first (block->headers, &c_nt2);
        while (node_header != NULL)
        {
          struct _ds_header_field *current_header =
            (struct _ds_header_field *) node_header->ptr;
          snprintf (heading, sizeof (heading),
                    "%s: %s\n", current_header->heading,
                    current_header->data);
          buffer_cat (header, heading);
          node_header = c_nt_next (block->headers, &c_nt2);
        }

        decode = block->body->data;

        /* Accumulate the bodies */
        if ((block->encoding == EN_BASE64
             || block->encoding == EN_QUOTED_PRINTABLE) && (!is_signed
                                                            || block->
                                                            original_signed_body
                                                            == NULL))
        {
          struct _ds_header_field *field;
          int is_attachment = 0;
          struct nt_node *node_hnt;
          struct nt_c c_hnt;

          node_hnt = c_nt_first (block->headers, &c_hnt);
          while (node_hnt != NULL)
          {
            field = (struct _ds_header_field *) node_hnt->ptr;
            if (field != NULL && field->heading != NULL
                && field->data != NULL)
              if (!strncasecmp (field->heading, "Content-Disposition", 19))
                if (!strncasecmp (field->data, "attachment", 10))
                  is_attachment = 1;
            node_hnt = c_nt_next (block->headers, &c_hnt);
          }


          if (!is_attachment)
          {
            LOGDEBUG ("decoding message block from encoding type %d",
                      block->encoding);
            decode = _ds_decode_block (block);
          }
        }

        if (decode != NULL)
        {
          char *decode2 = strdup(decode);

          /* -- PREFORMATTING BEGIN -- */

          // Hexadecimal decoding
          if (block->encoding == EN_8BIT) {
            char hex[5];
            int conv;
                                                                                
            strcpy(hex, "0x00");
                                                                                
            x = strchr(decode2, '%');
            while(x != NULL) {
              if (isxdigit((int) x[1]) && isxdigit((int) x[2])) {
                hex[2] = x[1];
                hex[3] = x[2];
                                                                                
                conv = strtol(hex, NULL, 16);
                                                                                
                x[0] = conv;
                memcpy(x+1, x+3, strlen(x+3));
              }
              x = strchr(x+1, '%');
            }
          }

          if (block->media_subtype == MST_HTML) {

            /* Remove long HTML Comments (Including DSPAM statistics) */
            x = strstr (decode2, "<!--");
            while (x != NULL)
            {
              y = strstr (x, "-->");
              if (y != NULL)
              {
                memcpy (x, y + 3, strlen (y + 3) + 1);
                x = strstr (x, "<!--");
              }
              else
              {
                x = strstr (x + 4, "<!--");
              }
            }
                                                                                
            /* Remove short HTML Comments */
            x = strstr (decode2, "<!");
            while (x != NULL)
            {
              y = strchr (x, '>');
              if (y != NULL)
              {
                memcpy (x, y + 1, strlen (y + 1) + 1);
                x = strstr (x, "<!");
              }
              else
              {
                x = strstr (x + 2, "<!");
              }
            }

            /* Remove short html tags and useless tags */
            x = strchr (decode2, '<');
            while (x != NULL)
            {
              y = strchr (x, '>');
              if (y != NULL
                  && (y - x <= 15
                      || !strncasecmp (x + 1, "td ", 3)
                      || !strncasecmp (x + 1, "!doctype", 8)
                      || !strncasecmp (x + 1, "blockquote", 10)
                      || !strncasecmp (x + 1, "table ", 6)
                      || !strncasecmp (x + 1, "tr ", 3)
                      || !strncasecmp (x + 1, "div ", 4)
                      || !strncasecmp (x + 1, "p ", 2)
                      || !strncasecmp (x + 1, "body ", 5) || !strchr (x, ' ')
                      || strchr (x, ' ') > y))
              {
                memcpy (x, y + 1, strlen (y + 1) + 1);
                x = strstr (x, "<");
              }
              else
              {
                x = strstr (x + 1, "<");
              }
            }
                                                                                
            /* -- PREFORMATTING END -- */
          }

          buffer_cat (body, decode2);
          free(decode2);
          if (decode != block->body->data)
          {
            if (is_signed)
            {
              LOGDEBUG
                ("message is signed.  retaining original text for reassembly");
              block->original_signed_body = block->body;
            }
            else
            {
              /* Switch the encoding */
              block->encoding = EN_7BIT;
              buffer_destroy (block->body);
            }

            block->body = buffer_create (decode);
            free (decode);

            if (!is_signed)
            {
              /* Rewrite the actual header */
              node_header = c_nt_first (block->headers, &c_nt2);
              while (node_header != NULL)
              {
                struct _ds_header_field
                  *header = (struct _ds_header_field *) node_header->ptr;
                if (!strcasecmp
                    (header->heading, "Content-Transfer-Encoding"))
                {
                  free (header->data);
                  header->data = strdup ("7bit");
                }
                node_header = c_nt_next (block->headers, &c_nt2);
              }
            }
          }
        }
      }
      node_nt = c_nt_next (CTX->message->components, &c_nt);
      i++;
    }

    if (CTX->flags & DSF_IGNOREHEADER)
      buffer_clear (header);

    if (header->data == NULL)
      buffer_cat (header, " ");

    if (body->data == NULL)
      buffer_cat (body, " ");
  }

  CTX->result = -1;

  spam_result = _ds_operate (CTX, header->data, body->data);

  if (spam_result == DSR_ISSPAM || spam_result == DSR_ISINNOCENT)
  {
    if (CTX->mode == DSM_FALSEPOSITIVE)
      spam_result = DSR_ISINNOCENT;
    else if (CTX->mode == DSM_ADDSPAM)
      spam_result = DSR_ISSPAM;
  }

  buffer_destroy (header);
  buffer_destroy (body);

  CTX->result = spam_result;

  if (CTX->result == DSR_ISSPAM || CTX->result == DSR_ISINNOCENT)
    return 0;
  else
  {
    LOGDEBUG ("received invalid result (! DSR_ISSPAM || DSR_INNOCENT): %d\n",
              CTX->result);
    return EUNKNOWN;
  }
}

/* dspam_getsource: returns the source ip (sender) for the message
   parameters:	DSPAM_CTX *CTX		pointer to context
		char *buf		pointer to buffer
		size_t size		size of buffer
   returns:	0 on success

   must be called _after_ the message is actualized (dspam_process)
*/

int
dspam_getsource (DSPAM_CTX * CTX, char *buf, size_t size,
                 const char *whitelist)
{
  struct _ds_message_block *current_block;
  struct _ds_header_field *current_heading = NULL;
  struct nt_node *node_nt;
  struct nt_c c;

  if (CTX->message == NULL)
    return EINVAL;

  node_nt = c_nt_first (CTX->message->components, &c);
  if (node_nt == NULL)
    return EINVAL;

  current_block = (struct _ds_message_block *) node_nt->ptr;

  node_nt = c_nt_first (current_block->headers, &c);
  while (node_nt != NULL)
  {
    current_heading = (struct _ds_header_field *) node_nt->ptr;
    if (!strcmp (current_heading->heading, "Received"))
    {
      char *data = strdup (current_heading->data);
      char *ptr = strstr (data, "from");
      char *tok;
      if (ptr != NULL)
      {
        tok = strtok (ptr, "[");
        if (tok != NULL)
        {
          int whitelisted = 0;
          FILE *file;

          tok = strtok (NULL, "]");

          if (tok != NULL)
          {
            if (!strcmp (tok, "127.0.0.1"))
              whitelisted = 1;
            if (whitelist != NULL)
            {
              file = fopen (whitelist, "r");
              if (file != NULL)
              {
                char buffer[128];
                while (fgets (buffer, sizeof (buffer), file) != NULL)
                {
                  chomp (buffer);
                  LOGDEBUG ("checking whitelist %s <> %s", tok, buffer);
                  if (!strcmp (tok, buffer))
                    whitelisted = 1;
                }
                fclose (file);
              }
            }

            if (!whitelisted)
            {
              strlcpy (buf, tok, size);
              free (data);
              return 0;
            }
          }
        }
      }
      free (data);
    }
    node_nt = c_nt_next (current_block->headers, &c);
  }

  return EFAILURE;
}

/* _ds_operate: operate on the message
                 - calculate the statistical probability the email is spam
                 - update tokens in dictionary according to result/mode
    parameters: DSPAM_CTX *CTX		pointer to context
                char *header		pointer to message header
                char *body		pointer to message body

       returns: DSR_ISSPAM		message is spam*
                DSR_ISINNOCENT		message is innocent*

       * returns result of actual calculation regardless of mode
*/

int
_ds_operate (DSPAM_CTX * CTX, char *headers, char *body)
{
  char *token;                  /* current token */
  char joined_token[32];        /* used for de-obfuscating tokens */
  char *previous_token = NULL;  /* used for chained tokens */

  char *line = NULL;            /* header broken up into lines */
  char *url_body;               /* urls broken up */

  char heading[128];            /* current heading */
  int alloc_joined = 0;         /* track joined token free()'s */
  int i = 0;

  /* Bayesian Calculations */
#if defined(BAYESIAN) || defined(DEBUG)
  float bay_top = 0.0;          /* Standard Bayesian */
  float bay_bot = 0.0;
  float bay_result = -1;
  long bay_used = 0;            /* Total tokens used in bayes */
#endif
#if defined(ALT_BAYESIAN) || defined(DEBUG)
  double abay_top = 0.0;        /* Alternative Bayesian */
  double abay_bot = 0.0;
  double abay_result = -1;
  long abay_used = 0;           /* Total tokens used in altbayes */
#endif

  /* Chi-Square used to calculate confidence */
  float chi_top = 0.0;          /* Chi-Square */
  float chi_bot = 0.0;
  float chi_result = -1;
  double p = 0.0, q = 0.0, s = 0.0;     /* Chi-Calculations */
  long chi_used = 0;            /* Total tokens used in chi */

  /* Long Hashed Token Tree: Track tokens, frequencies, and stats */
  struct lht *freq = lht_create (1543);
  struct lht_node *node_lht;
  struct lht_c c_lht;
  struct _ds_spam_stat stat;

  struct tbt *index = tbt_create ();    /* Binary tree index */
  struct tbt_node *node_tbt;

  struct nt *header = NULL;     /* header array */
  struct nt_node *node_nt;
  struct nt_c c_nt;
  long body_length; 
  int spammy = 0;

  int signature_null = (CTX->signature == NULL) ? 1 : 0;

#ifdef AUTO_WHITELIST
  unsigned long long whitelist_token = 0;
  int do_whitelist = 0;
#endif

  if (body == NULL)
    body_length = 0;
  else
    body_length = strlen(body);

  joined_token[0] = 0;

  if (freq == NULL || index == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    goto bail_unknown;
  }

  CTX->result = (CTX->mode == DSM_ADDSPAM) ? DSR_ISSPAM : DSR_ISINNOCENT;

  if (CTX->flags & DSF_SIGNATURE && CTX->mode == DSM_CLASSIFY && !signature_null)
  {
    int num_tokens =
      CTX->signature->length / sizeof (struct _ds_signature_token);
    struct _ds_signature_token t;

    for (i = 0; i < num_tokens; i++)
    {
      char x[128];
      memcpy (&t,
              (char *) CTX->signature->data +
              (i * sizeof (struct _ds_signature_token)),
              sizeof (struct _ds_signature_token));
      snprintf (x, sizeof (x), "E: %llu", t.token);
      lht_hit (freq, t.token, x);
      lht_setfrequency (freq, t.token, t.frequency);
    }
  }
  else
  {

    header = nt_create (NT_CHAR);
    if (header == NULL)
    {
      LOG (LOG_CRIT, ERROR_MEM_ALLOC);
      goto bail_unknown;
    }

    /* HEADER: Split up the text into tokens, include heading */
    line = strtok (headers, "\n");
    while (line != NULL)
    {
      nt_add (header, line);
      line = strtok (NULL, "\n");
    }

    node_nt = c_nt_first (header, &c_nt);
    heading[0] = 0;
    while (node_nt != NULL)
    {
      int is_received;
      joined_token[0] = 0;
      alloc_joined = 0;

      line = node_nt->ptr;
      token = strtok (line, ":");
      if (token != NULL && token[0] != 32 && token[0] != 9
          && !strstr (token, " "))
      {
        strlcpy (heading, token, 128);
        previous_token = NULL;
      }

#ifdef VERBOSE
      LOGDEBUG ("Reading '%s' header", heading);
#endif

#ifdef AUTO_WHITELIST
      /* Track the entire From: line for auto-whitelisting */

      if (!strcmp(heading, "From")) {
        char wl[256];
        char *fromline = line + 5;

        if (fromline[0] == 32) 
          fromline++;

        snprintf(wl, sizeof(wl), "%s*%s", heading, fromline);
        whitelist_token = _ds_getcrc64(wl); 
        lht_hit (freq, whitelist_token, wl);
      }
#endif

      is_received = (!strcmp (heading, "Received") ? 1 : 0);

      if (is_received)
        token = strtok (NULL, DELIMITERS_HEADING);
      else
        token = strtok (NULL, DELIMITERS);

      while (token != NULL)
      {
        int l;

        l = strlen (token);
        if ((l > 2 && l < 25)
            || (l == 2 && !strchr (token, '$') && !strchr (token, '!')))
        {

#ifdef VERBOSE
          LOGDEBUG ("Processing '%s' token in '%s' header", token, heading);
#endif

          /* If we had to join a token together (e.g. S E X), process it */
          if (joined_token[0] != 0)
          {
            if (strlen (joined_token) < 25 && joined_token[1] != 0)
            {
              if (!_ds_process_header_token
                  (CTX, joined_token,
                   previous_token, freq, heading)
                  && (CTX->flags & DSF_CHAINED))
              {
                alloc_joined = 1;
                previous_token = strdup (joined_token);
              }
            }
            joined_token[0] = 0;
          }

          if (!_ds_process_header_token
              (CTX, token, previous_token, freq,
               heading) && (CTX->flags & DSF_CHAINED))
          {
            if (alloc_joined)
            {
              free (previous_token);
              alloc_joined = 0;
            }
            previous_token = token;
          }

        }
        else if (l == 1
                 || (l == 2 && (strchr (token, '$') || strchr (token, '!'))))
        {
          strlcat (joined_token, token, sizeof (joined_token));
        }

        if (is_received)
          token = strtok (NULL, DELIMITERS_HEADING);
        else
          token = strtok (NULL, DELIMITERS);
      }
      node_nt = c_nt_next (header, &c_nt);

      if (joined_token[0] != 0)
      {
        if (strlen (joined_token) < 25 && joined_token[1] != 0)
        {
          _ds_process_header_token (CTX, joined_token,
                                    previous_token, freq, heading);
        }
      }

    }

    nt_destroy (header);

    if (alloc_joined)
      free (previous_token);

    previous_token = NULL;

    /* BODY: Split up URLs into tokens, count frequency */
    url_body = strdup (body);
    if (url_body != NULL)
    {
      char combined_token[256];
      char *url_ptr = url_body;
      int url_length;
      unsigned long long crc;

      token = strstr (url_ptr, "http://");
      while (token != NULL)
      {
        url_ptr = token;

        token = strtok (token, " \n\">");
        if (token != NULL)
        {
          url_length = strlen (token);

          /* Individual tokens form the URL */
          token = strtok (token, DELIMITERS);
          while (token != NULL)
          {
            snprintf (combined_token, sizeof (combined_token), "Url*%s",
                      token);
            crc = _ds_getcrc64 (combined_token);
            lht_hit (freq, crc, combined_token);
            token = strtok (NULL, DELIMITERS);
          }

          memset (body + (url_ptr - url_body), 32, url_length);
          url_ptr += url_length + 1;
          token = strstr (url_ptr, "http://");
        }
        else
          token = NULL;
      }
      free (url_body);
    }

    url_body = strdup (body);
    if (url_body != NULL)
    {
      char combined_token[256];
      char *url_ptr = url_body;
      int url_length;
      unsigned long long crc;

      url_ptr = url_body;
      token = strstr (url_ptr, "href=\"");
      while (token != NULL)
      {
        url_ptr = token + 6;

        token = strtok (url_ptr, " \n\">");
        if (token != NULL)
        {
          url_length = strlen (token);

          /* Individual tokens form the URL */
          token = strtok (url_ptr, DELIMITERS);
          while (token != NULL)
          {
            snprintf (combined_token, sizeof (combined_token), "Url*%s",
                      token);
            crc = _ds_getcrc64 (combined_token);
            lht_hit (freq, crc, combined_token);
            token = strtok (NULL, DELIMITERS);
          }

          memset (body + (url_ptr - url_body), 32, url_length);

          url_ptr += url_length + 1;
          token = strstr (url_ptr, "href=\"");
        }
        else
          token = NULL;
      }

      free (url_body);
    }

    /* BODY: Split up the text into tokens, count frequency */
    joined_token[0] = 0;
    alloc_joined = 0;
    token = strtok (body, DELIMITERS);
    while (token != NULL)
    {
      int l = strlen (token);
      if ((l > 2 && l < 25)
          || (l == 2 && !strchr (token, '$') && !strchr (token, '!')))
      {
        /* If we had to join a token together (e.g. S E X), process it */
        if (joined_token[0] != 0)
        {
          if (strlen (joined_token) < 25 && joined_token[1] != 0)
          {
            if (!_ds_process_body_token
                (CTX, joined_token,
                 previous_token, freq) && (CTX->flags & DSF_CHAINED))
            {
              alloc_joined = 1;
              previous_token = strdup (joined_token);
            }
          }
          joined_token[0] = 0;
        }

        if (!_ds_process_body_token
            (CTX, token, previous_token, freq) && (CTX->flags & DSF_CHAINED))
        {
          if (alloc_joined)
          {
            alloc_joined = 0;
            free (previous_token);
          }
          previous_token = token;
        }
      }
      else if (l == 1
               || (l == 2 && (strchr (token, '$') || strchr (token, '!'))))
      {
        strlcat (joined_token, token, sizeof (joined_token));
      }
      token = strtok (NULL, DELIMITERS);
    }

    if (joined_token[0] != 0)
    {
      if (strlen (joined_token) < 25 && joined_token[1] != 0)
      {
        _ds_process_body_token (CTX, joined_token, previous_token, freq);
      }
    }
  }

  /* Load all token statistics */
  if (_ds_getall_spamrecords (CTX, freq))
  {
    LOGDEBUG ("_ds_getall_spamrecords() failed");
    return EUNKNOWN;
  }

  /* Bayesian Noise Reduction - Statistical Feature Omission Logic
     http://www.nuclearelephant.com/projects/dspam/bnr.html */

  i = 0;
#define DUB_MIN_THRESHHOLD	4

  if (CTX->flags & DSF_NOISE && 
       ((CTX->mode == DSM_CLASSIFY || CTX->mode == DSM_PROCESS) 
    && (!(CTX->flags & DSF_CORPUS)) && 
       (CTX->totals.total_innocent + CTX->totals.innocent_corpusfed)>2500))
  {
    float previous = 0.000000; /* Previous token probability */
    struct lht_node *node_lht2 = NULL;
    int total_noise = 0, total_clear = 0;
    int dub = 0; /* Dub strings mode */
    struct nt_node *dub_start = NULL, *dub_end = NULL;
    int dub_length = 0;

    node_nt = c_nt_first(freq->order, &c_nt);
    while(node_nt != NULL) {
      node_lht = (struct lht_node *) node_nt->ptr;
      if (node_nt->next != NULL)
        node_lht2 = (struct lht_node *) node_nt->next->ptr;
      else
        node_lht2 = node_lht;

      if (previous == 0.000000)
        _ds_calc_stat (CTX, node_lht->key, &node_lht->s);
      _ds_calc_stat (CTX, node_lht2->key, &node_lht2->s);

      if (node_lht->s.probability >= 0.8)
        spammy++;

      /* Turn off dub if we come to a noise-ending marker */
      if (dub && ((node_lht->s.probability >= 0.8) 
        || (node_lht->s.probability <= 0.25 && node_lht2->s.probability <= 0.25)))
      {
        dub = 0;
        dub_end = node_nt;

        /* Undub if the dub length is too short */
        if (dub_length<DUB_MIN_THRESHHOLD && dub_length>1) {
          struct nt_node *undub_this;

          undub_this = dub_start->next;
          while(undub_this != dub_end) {
            ((struct lht_node *) undub_this->ptr)->frequency++;
            undub_this = undub_this->next;
          }
        }
      }

      if (dub)
        dub_length++;

      /* Identify noise, turn on dubbing */
      if (dub /* Dub is on OR */
         || (!strchr(node_lht->token_name, '*') /* Not a header */ 
         && (node_lht->s.probability <= 0.30 || node_lht->s.probability >= 0.90)
         && ((previous >= 0.40 && previous <= 0.60) ||
             (node_lht2->s.probability >= 0.40 
           && node_lht2->s.probability <= 0.60)))) 
      {
        i++;
        node_lht->frequency-=2;

        total_noise++;

        if (node_lht->s.probability <= 0.30) {

          if  (dub && !strchr(node_lht->token_name, '*') /* Not a header */
           && (node_lht->s.probability <= 0.30)
           && ((previous >= 0.40 && previous <= 0.60) ||
             (node_lht2->s.probability >= 0.40
           && node_lht2->s.probability <= 0.60)))
          {
            dub_start = node_nt;
          }

          if (!dub) {
            dub_length = 1;
            dub_start = node_nt;
            dub = 1;
          }
        }
      } else { 
        total_clear++;
      }

      previous = node_lht->s.probability;
      node_nt = c_nt_next(freq->order, &c_nt);
    }
    dub = 0;
    dub_start = NULL;
    dub_end = NULL;

    /* Do the same for chained tokens; duplicate code incase we want to
       treat chained tokens differently down the road */
    previous = 0.000000;
    node_nt = c_nt_first(freq->chained_order, &c_nt);
    while(node_nt != NULL) {
      node_lht = (struct lht_node *) node_nt->ptr;
      if (node_nt->next != NULL)
        node_lht2 = (struct lht_node *) node_nt->next->ptr;
      else
        node_lht2 = node_lht;
                                                                                
      if (previous == 0.000000)
        _ds_calc_stat (CTX, node_lht->key, &node_lht->s);
      _ds_calc_stat (CTX, node_lht2->key, &node_lht2->s);

     if (node_lht->s.probability >= 0.8)
        spammy++;
                                                                                
      /* Turn off dub if we come to a noise-ending marker */
      if (dub && ((node_lht->s.probability >= 0.8)
        || (node_lht->s.probability <= 0.25 && node_lht2->s.probability <= 0.25)))
      {
        dub = 0;
        dub_end = node_nt;
                                                                                
        /* Undub if the dub length is too short */
        if (dub_length<DUB_MIN_THRESHHOLD && dub_length>1) {
          struct nt_node *undub_this;
                                                                                
          undub_this = dub_start->next;
          while(undub_this != dub_end) {
            ((struct lht_node *) undub_this->ptr)->frequency++;
            undub_this = undub_this->next;
          }
        }
      }

      if (dub)
        dub_length++;

      /* Identify noise, turn on dubbing */
      if (dub /* Dub is on OR */
         || (!strchr(node_lht->token_name, '*') /* Not a header */
         && (node_lht->s.probability <= 0.30) 
            && ((previous >= 0.40 && previous <= 0.60) ||
              (node_lht2->s.probability >= 0.40
            && node_lht2->s.probability <= 0.60))))
      {
        i++;
        node_lht->frequency-=2;
        total_noise++;

        if (node_lht->s.probability <= 0.30) {

         if  (dub && !strchr(node_lht->token_name, '*') /* Not a header */
           && (node_lht->s.probability <= 0.30)
           && ((previous >= 0.40 && previous <= 0.60) ||
             (node_lht2->s.probability >= 0.40
           && node_lht2->s.probability <= 0.60)))
          {
            dub_start = node_nt;
          }

          if (!dub) {
            dub_length = 1;
            dub_start = node_nt;
            dub = 1;
          }
        }
      } else {
        total_clear++;
      }
                                                                                
      previous = node_lht->s.probability;
      node_nt = c_nt_next(freq->chained_order, &c_nt);
    }

    /* Check and make sure our noise ratio is within range */
    previous = 100.0*((total_noise+0.0)/(total_clear+total_noise));
    if ((previous > 77.5 && body_length>3500) 
     || (previous > 93.0 && body_length<=3500))
      i = 0;

    LOGDEBUG("BNR noise ratio: %2.2f Spammy Tokens: %d Filtered: %d\n", previous, spammy, i);
  } else {
    i = -1;
  }

/*
  {
    printf("DUBBING\n\n");
    node_nt = c_nt_first(freq->order, &c_nt);
    while(node_nt != NULL) {
      node_lht = (struct lht_node *)node_nt->ptr;
      if (node_lht->frequency<=0)
        printf("%s ", node_lht->token_name);
        node_nt = node_nt->next;
    }                                                                                                                                                  
    printf("\n\n");

    node_nt = c_nt_first(freq->chained_order, &c_nt);
     while(node_nt != NULL) {
     node_lht = (struct lht_node *)node_nt->ptr;
     if (node_lht->frequency<=0)
        printf("%s ", node_lht->token_name);
      node_nt = node_nt->next;
    }                                                                                                                                                  
    printf("\n\n");
  }


  {
    printf("---------\nLEFT WITH\n\n");
    node_nt = c_nt_first(freq->order, &c_nt);
    while(node_nt != NULL) {
      node_lht = (struct lht_node *)node_nt->ptr;
      if (node_lht->frequency>0)
        printf("%s ", node_lht->token_name);
      node_nt = node_nt->next;
    }

    printf("\n\n");

    node_nt = c_nt_first(freq->chained_order, &c_nt);
    while(node_nt != NULL) {
      node_lht = (struct lht_node *)node_nt->ptr;
      if (node_lht->frequency>0)
        printf("%s ", node_lht->token_name);
      node_nt = node_nt->next;
    }

    printf("\n\n");
  }

  printf("SPAMMY: %d\n\n", spammy);
*/

  /* Create a binary tree index sorted by a token's delta from .5 */
  node_lht = c_lht_first (freq, &c_lht);
  while (node_lht != NULL)
  {
    if (i == -1) 
      _ds_calc_stat (CTX, node_lht->key, &node_lht->s);

#ifdef AUTO_WHITELIST
    if (node_lht->key == whitelist_token 
     && node_lht->s.spam_hits == 0
     && node_lht->s.innocent_hits > 10
     && (CTX->mode == DSM_PROCESS || CTX->mode == DSM_CLASSIFY) &&
        (! (CTX->flags & DSF_CORPUS))) {
       do_whitelist = 1;
     }
#endif
    if (node_lht->frequency>0 || i<5 || (spammy < 10 && body_length>=3500) ||
        (spammy < 5 && body_length <= 3500)) {
      tbt_add (index, node_lht->s.probability, node_lht->key,
             node_lht->frequency, _ds_compute_complexity(node_lht->token_name));
    }

#ifdef VERBOSE
    LOGDEBUG ("Token: %s [%f]", node_lht->token_name,
              node_lht->s.probability);
#endif

    node_lht = c_lht_next (freq, &c_lht);
  }

  /* Take the 15 most interesting tokens and generate a score, update dictionary */

  if (index->items == 0)
  {
    LOGDEBUG ("no tokens found in message");
    tbt_destroy (index);
    lht_destroy (freq);
    return EINVAL;
  }

  /* Initialize signature if requested */
  if (CTX->flags & DSF_SIGNATURE && (CTX->mode != DSM_CLASSIFY || signature_null))
  {
    CTX->signature = calloc (1, sizeof (struct _ds_spam_signature));
    if (CTX->signature == NULL)
    {
      LOG (LOG_CRIT, "memory allocation error");
      tbt_destroy (index);
      lht_destroy (freq);
      return EUNKNOWN;
    }

    CTX->signature->length =
      sizeof (struct _ds_signature_token) * freq->items;
    CTX->signature->data = malloc (CTX->signature->length);
    if (CTX->signature->data == NULL)
    {
      LOG (LOG_CRIT, "memory allocation error");
      free (CTX->signature);
      CTX->signature = NULL;
      tbt_destroy (index);
      lht_destroy (freq);
      return EUNKNOWN;
    }
  }

  node_tbt = tbt_first (index);

  for (i = 0; i < ((index->items > 27) ? 27 : index->items); i++)
  {
    unsigned long long crc;
    char *token_name;

    crc = node_tbt->token;
    token_name = lht_gettoken (freq, crc);

    if (lht_getspamstat (freq, crc, &stat) || token_name == NULL)
    {
      node_tbt = tbt_next (node_tbt);
      continue;
    }

    if (CTX->flags & DSF_CORPUS && CTX->mode != DSM_CLASSIFY)
    {
      if (CTX->mode == DSM_ADDSPAM)
        stat.probability = 1.00;
      else
        stat.probability = 0.00;
    }

    /* BEGIN Combine Token Values */

    /* Bayesian */
#if defined(BAYESIAN) || defined(DEBUG)
    if (bay_used < 15)
    {

#ifdef BAYESIAN
#ifndef VERBOSE
      if (CTX->mode != DSM_CLASSIFY)
      {
#endif
        LOGDEBUG ("[bayes] combining: [%2.6f] %s (%df, %lds, %ldi)",
                  stat.probability, token_name, lht_getfrequency (freq, crc),
                  stat.spam_hits, stat.innocent_hits);
#ifndef VERBOSE
      }
#endif
#endif

      if (bay_used == 0)
      {
        bay_top = stat.probability;
        bay_bot = 1 - stat.probability;
      }
      else
      {
        bay_top *= stat.probability;
        bay_bot *= (1 - stat.probability);
      }

      bay_used++;
    }
#endif

    /* Alternative Bayesian */
#if defined(ALT_BAYESIAN) || defined(DEBUG)
    if (abay_used < 27)
    {
#ifdef ALT_BAYESIAN
      if (CTX->mode != DSM_CLASSIFY)
      {
        LOGDEBUG ("[altbayes] combining: [%2.6f] %s (%df, %lds, %ldi)",
                  stat.probability, token_name, lht_getfrequency (freq, crc),
                  stat.spam_hits, stat.innocent_hits);
      }
#endif

      if (abay_used == 0)
      {
        abay_top = stat.probability;
        abay_bot = (1 - stat.probability);
      }
      else
      {
        abay_top *= stat.probability;
        abay_bot *= (1 - stat.probability);
      }

      abay_used++;

      if (abay_used < 27 && lht_getfrequency (freq, crc) > 1)
      {
#ifdef ALT_BAYESIAN
        if (CTX->mode != DSM_CLASSIFY)
        {
          LOGDEBUG ("[altbayes] combining: [%2.6f] %s (%df, %lds, %ldi)",
                    stat.probability, token_name, lht_getfrequency (freq,
                                                                    crc),
                    stat.spam_hits, stat.innocent_hits);
        }
#endif
        abay_used++;
        abay_top *= stat.probability;
        abay_bot *= (1 - stat.probability);
      }
    }
#endif

    /* Chi-Square */

    /* Definitions used in BogoFilter */
#define CHI_S	0.010           /* Sensitivity */
#define CHI_X	0.415           /* Value to use when N = 0 */
#define CHI_CUTOFF	0.54

    if (chi_used < 25)
    {
      float probability;
      long n = (index->items > 25) ? 25 : index->items;

      probability = ((CHI_S * CHI_X) + (n * stat.probability)) / (CHI_S + n);

#ifdef CHI_SQUARE
      if (CTX->mode != DSM_CLASSIFY)
      {
        LOGDEBUG ("[chi-sq] combining: [%2.6f] %s (%df, %lds, %ldi)",
                  stat.probability, token_name, lht_getfrequency (freq, crc),
                  stat.spam_hits, stat.innocent_hits);
      }
#endif

      if (probability < 0.3 || probability > 0.7)
      {

        if (chi_used == 0)
        {
          chi_top = probability;
          chi_bot = (1 - probability);
        }
        else
        {
          chi_top *= probability;
          chi_bot *= (1 - probability);
        }

        chi_used++;

        if (chi_used < 25 && lht_getfrequency (freq, crc) > 1)
        {
#ifdef CHI_SQUARE
          if (CTX->mode != DSM_CLASSIFY)
          {
            LOGDEBUG ("[chi-sq] combining: [%2.6f] %s (%df, %lds, %ldi)",
                      stat.probability, token_name, lht_getfrequency (freq,
                                                                      crc),
                      stat.spam_hits, stat.innocent_hits);

          }
#endif
          chi_used++;
          chi_top *= probability;
          chi_bot *= (1 - probability);
        }
      }
    }

/* END Combine Token Values */

    node_tbt = tbt_next (node_tbt);
  }

/* BEGIN Calculate Individual Probabilities */

#if defined(BAYESIAN) || defined(DEBUG)
  bay_result = (bay_top) / (bay_top + bay_bot);
  LOGDEBUG ("Bayes Probability: %f Samples: %ld", bay_result, bay_used);
#endif

#if defined(ALT_BAYESIAN) || defined(DEBUG)
  abay_result = (abay_top) / (abay_top + abay_bot);
  LOGDEBUG ("Alternative Bayes Probability: %f Samples: %ld", abay_result,
            abay_used);
#endif

  /* Chi-Square */
  if (chi_used == 0)
  {
    p = q = s = 0;
  }
  else
  {
    p = 1.0 - pow (chi_bot, 1.0 / chi_used);
    q = 1.0 - pow (chi_top, 1.0 / chi_used);
    s = (p - q) / (p + q);
    s = (s + 1.0) / 2.0;
  }

  chi_result = s;

#if defined(CHI_SQUARE) || defined(DEBUG)
  LOGDEBUG
    ("Chi-Square Confidence: %f (Spamminess: %f, Non-Spamminess: %f, Samples: %ld)",
     chi_result, p, q, chi_used);
#endif

/* END Calculate Individual Probabilities */

/* BEGIN Determine Result */

#ifdef BAYESIAN
  if (bay_result >= 0.9 && CTX->mode != DSM_FALSEPOSITIVE)
  {
    CTX->result = DSR_ISSPAM;
    CTX->probability = bay_result;
  }
#endif

#ifdef ALT_BAYESIAN
  if (abay_result >= 0.9 && CTX->mode != DSM_FALSEPOSITIVE)
  {
    CTX->result = DSR_ISSPAM;
    CTX->probability = abay_result;
  }
#endif

#ifdef CHI_SQUARE
  if (chi_result >= CHI_CUTOFF && CTX->mode != DSM_FALSEPOSITIVE)
  {
    CTX->result = DSR_ISSPAM;
    if (CTX->probability < 0)
      CTX->probability = chi_result;
  }
#endif

#ifdef AUTO_WHITELIST
  if (do_whitelist) {
    LOGDEBUG("auto-whitelisting this message");
    CTX->result = DSR_ISINNOCENT;
 }
#endif

  /* If the message is innocent, we'll need to assign one probability */
  if (CTX->probability < 0)
  {
#ifdef BAYESIAN
    CTX->probability = bay_result;
#else
#ifdef ALT_BAYESIAN
    CTX->probability = abay_result;
#else
    CTX->probability = chi_result;
#endif
#endif
  }

/* END Determine Result */

#ifdef DEBUG
  if (abay_result >= 0.9 && bay_result < 0.9)
  {
    LOGDEBUG ("CATCH: Alternative Bayesian");
  }
  else if (abay_result < 0.9 && bay_result >= 0.9)
  {
    LOGDEBUG ("MISS: Alternative Bayesian");
  }

  if (chi_result >= CHI_CUTOFF && bay_result < 0.9)
  {
    LOGDEBUG ("CATCH: Chi-Square");
  }
  else if (chi_result < CHI_CUTOFF && bay_result >= 0.9)
  {
    LOGDEBUG ("MISS: Chi-Square");
  }
#endif

  /* Update Totals */

  /* SPAM */
  if (CTX->result == DSR_ISSPAM && CTX->mode != DSM_CLASSIFY)
  {
    CTX->totals.total_spam++;

    if (CTX->mode == DSM_ADDSPAM)
    {

      if (CTX->flags & DSF_CORPUS || CTX->flags & DSF_INOCULATE)
        CTX->totals.spam_corpusfed++;
      else
      {
        CTX->totals.spam_misclassified++;
        CTX->totals.total_innocent -=
          (CTX->totals.total_innocent > 0) ? 1 : 0;
      }
    }

    /* INNOCENT */
  }
  else if (CTX->result == DSR_ISINNOCENT && CTX->mode != DSM_CLASSIFY)
  {
    CTX->totals.total_innocent++;

    if (CTX->flags & DSF_CORPUS || CTX->flags & DSF_INOCULATE)
    {
      CTX->totals.innocent_corpusfed++;
    }
    else if (CTX->mode == DSM_FALSEPOSITIVE)
    {
      CTX->totals.innocent_misclassified++;
      CTX->totals.total_spam -= (CTX->totals.total_spam > 0) ? 1 : 0;
    }
  }

  i = 0;
  node_lht = c_lht_first(freq, &c_lht);
  while (node_lht != NULL)
//HERE
  {
    unsigned long long crc;
    char *token_name;

    crc = node_lht->key;
    token_name = node_lht->token_name;

    if (CTX->flags & DSF_SIGNATURE && (CTX->mode != DSM_CLASSIFY || signature_null))
    {
      struct _ds_signature_token t;

      t.token = crc;
      t.frequency = lht_getfrequency (freq, t.token);
      memcpy ((char *) CTX->signature->data +
              (i * sizeof (struct _ds_signature_token)), &t,
              sizeof (struct _ds_signature_token));
    }

    if (CTX->flags & DSF_CORPUS)
    {
      if (CTX->mode == DSM_ADDSPAM)
        node_lht->s.probability = 1.00;
      else
        node_lht->s.probability = 0.00;
    }

    if (CTX->result == DSR_ISSPAM)
    {
      if (CTX->flags & DSF_INOCULATE)
      {
        if (node_lht->s.innocent_hits < 2 && node_lht->s.spam_hits < 5)
          node_lht->s.spam_hits += 5;
        else
          node_lht->s.spam_hits += 2;
      }
      else
      {
        node_lht->s.spam_hits++;
      }
    }

    else if (CTX->mode != DSM_FALSEPOSITIVE)
      node_lht->s.innocent_hits++;

    if (CTX->mode == DSM_ADDSPAM && (!(CTX->flags & DSF_CORPUS)))
      node_lht->s.innocent_hits--;

    if (CTX->mode == DSM_FALSEPOSITIVE)
    {
      node_lht->s.innocent_hits++;

      node_lht->s.spam_hits--;
      if (node_lht->s.spam_hits < 0)
        node_lht->s.spam_hits = 0;
    }

    if (node_lht->s.innocent_hits < 0)
      node_lht->s.innocent_hits = 0;

    i++;

    node_lht = c_lht_next(freq, &c_lht);
  }

  if (_ds_setall_spamrecords (CTX, freq))
  {
    LOGDEBUG ("_ds_setall_spamrecords() failed");
    return EUNKNOWN;
  }

  tbt_destroy (index);
  lht_destroy (freq);

  if (CTX->mode == DSM_FALSEPOSITIVE)
  {
    CTX->probability = 0.0;
    CTX->result = DSR_ISINNOCENT;
  }
  else if (CTX->mode == DSM_ADDSPAM)
  {
    CTX->probability = 1.0;
    CTX->result = DSR_ISSPAM;
  }

  if (CTX->result == DSR_ISSPAM)
  {
    CTX->confidence = chi_result;
  }
  else
  {
    CTX->confidence = 1.0 - chi_result;
  }

  return (CTX->result == DSR_ISSPAM) ? DSR_ISSPAM : DSR_ISINNOCENT;

bail_unknown:
  tbt_destroy (index);
  nt_destroy (header);
  lht_destroy (freq);
  return EUNKNOWN;
}

int
_ds_process_signature (DSPAM_CTX * CTX)
{
  struct _ds_signature_token t;
  int num_tokens, i;
  struct lht *freq = lht_create (1543);
  struct lht_node *node_lht;
  struct lht_c c_lht;

  if (CTX->signature == NULL)
  {
    LOGDEBUG ("DSF_SIGNATURE specified, but no signature provided.");
    return EINVAL;
  }

  if (freq == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  LOGDEBUG ("processing signature.  length: %ld", CTX->signature->length);

  CTX->result = -1;

  if (CTX->mode == DSM_FALSEPOSITIVE)
  {
    CTX->totals.innocent_misclassified++;
    CTX->totals.total_spam -= (CTX->totals.total_spam > 0) ? 1 : 0;
    CTX->totals.total_innocent++;

  }
  else if (CTX->mode == DSM_ADDSPAM)
  {
    if (CTX->flags & DSF_CORPUS || CTX->flags & DSF_INOCULATE)
      CTX->totals.spam_corpusfed++;
    else
    {
      CTX->totals.spam_misclassified++;
      CTX->totals.total_innocent -= (CTX->totals.total_innocent > 0) ? 1 : 0;
    }
    CTX->totals.total_spam++;

  }
  else if (CTX->mode != DSM_CLASSIFY)
  {

    /* If it's not a false positive, forwarded spam, or corpus-added spam,
     * it's a corpus-added innocent */

    if (CTX->flags & DSF_CORPUS || CTX->flags & DSF_INOCULATE)
      CTX->totals.innocent_corpusfed++;

    CTX->totals.total_innocent++;
  }

  num_tokens = CTX->signature->length / sizeof (struct _ds_signature_token);

  LOGDEBUG ("reversing %d tokens", num_tokens);
  for (i = 0; i < num_tokens; i++)
  {
    memcpy (&t,
            (char *) CTX->signature->data +
            (i * sizeof (struct _ds_signature_token)),
            sizeof (struct _ds_signature_token));
    lht_hit (freq, t.token, "-");
    lht_setfrequency (freq, t.token, t.frequency);
  }

  if (_ds_getall_spamrecords (CTX, freq))
    return EUNKNOWN;

  node_lht = c_lht_first (freq, &c_lht);
  while (node_lht != NULL)
  {
    if (CTX->mode == DSM_FALSEPOSITIVE)
    {
      node_lht->s.innocent_hits++;
      node_lht->s.spam_hits -= (node_lht->s.spam_hits > 0) ? 1 : 0;
    }
    else
    {
      if (!(CTX->flags & DSF_CORPUS))
        node_lht->s.innocent_hits -= (node_lht->s.innocent_hits > 0) ? 1 : 0;
      if (CTX->flags & DSF_INOCULATE)
      {
        if (node_lht->s.innocent_hits < 2 && node_lht->s.spam_hits < 5)
          node_lht->s.spam_hits += 5;
        else
          node_lht->s.spam_hits += 2;
      }
      else
      {
        node_lht->s.spam_hits++;
      }
    }
    node_lht = c_lht_next (freq, &c_lht);
  }

  if (_ds_setall_spamrecords (CTX, freq))
    return EUNKNOWN;

  if (CTX->mode == DSM_FALSEPOSITIVE)
  {
    CTX->probability = 0.0;
    CTX->result = DSR_ISINNOCENT;
  }
  else
  {
    CTX->probability = 1.0;
    CTX->result = DSR_ISSPAM;
  }

  return 0;
}

/* load/save spam statistics */

int
_ds_calc_stat (DSPAM_CTX * CTX, unsigned long long token,
               struct _ds_spam_stat *s)
{

  int min_hits = 5;
                                                                                
  if (CTX->totals.total_innocent + CTX->totals.innocent_corpusfed < 1000
  && CTX->totals.total_innocent < CTX->totals.total_spam)
    min_hits = 10+((CTX->totals.total_spam-CTX->totals.total_innocent)/20);

  if ((CTX->totals.total_innocent + CTX->totals.innocent_corpusfed) < 2500  && 
      (CTX->totals.total_innocent + CTX->totals.innocent_corpusfed) >= 1000 &&
        (CTX->totals.total_spam + CTX->totals.spam_corpusfed) > 
        (CTX->totals.total_innocent + CTX->totals.innocent_corpusfed))
  {
    float spams =
      (CTX->totals.total_spam * 1.0 /
       (CTX->totals.total_spam * 1.0 +
        CTX->totals.total_innocent * 1.0)) * 100;
    min_hits = 10+(spams/2);
  }

  if (min_hits < 5)
    min_hits = 5;

#ifdef TUM
  if (min_hits>20)
    min_hits = 20;
#endif

  if (CTX->mode == DSM_ADDSPAM)
    s->probability = .7;
  else
    s->probability = .4;

  if (CTX->totals.total_spam > 0 && CTX->totals.total_innocent > 0 &&
      ((s->spam_hits * 1.0 / CTX->totals.total_spam * 1.0) +
       (s->innocent_hits * 1.0 / CTX->totals.total_innocent * 1.0)) > 0)
  {
#ifdef BIAS
    s->probability =
      (s->spam_hits * 1.0 / CTX->totals.total_spam * 1.0) /
      ((s->spam_hits * 1.0 / CTX->totals.total_spam * 1.0) +
       (s->innocent_hits * 2.0 / CTX->totals.total_innocent * 1.0));
#else
    s->probability =
      (s->spam_hits * 1.0 / CTX->totals.total_spam * 1.0) /
      ((s->spam_hits * 1.0 / CTX->totals.total_spam * 1.0) +
       (s->innocent_hits * 1.0 / CTX->totals.total_innocent * 1.0));
#endif
  }

  if (s->innocent_hits < 0)
    s->innocent_hits = 0;

  if (s->spam_hits < 0)
    s->spam_hits = 0;

  if (s->spam_hits == 0 && s->innocent_hits > 0)
  {
    if (s->innocent_hits > 50)
      s->probability = 0.0060;
    else if (s->innocent_hits > 10)
      s->probability = 0.0001;
    else
      s->probability = 0.0002;
  }
  else if (s->spam_hits > 0 && s->innocent_hits == 0)
  {
    if (s->spam_hits >= 10)
      s->probability = 0.9999;
    else
      s->probability = 0.9998;
  }

#ifdef BIAS
  if (s->spam_hits + (2 * s->innocent_hits) < min_hits
      || CTX->totals.total_innocent < min_hits)
#else
  if (s->spam_hits + s->innocent_hits < min_hits
      || CTX->totals.total_innocent < min_hits)
#endif
    s->probability = .4;

  if (s->probability < 0.0001)
    s->probability = 0.0001;

  if (s->probability > 0.9999)
    s->probability = 0.9999;

  return 0;
}

int
_ds_process_header_token (DSPAM_CTX * CTX, char *token,
                          const char *previous_token, struct lht *freq,
                          const char *heading)
{
  int all_num = 1, i;
  char combined_token[256];
  int len = 0;
  int is_received;
  unsigned long long crc;

  is_received = (!strcmp (heading, "Received") ? 1 : 0);

  if (is_received && strlen (token) < 6)
    return EINVAL;

  for (i = 0; token[i] != 0; i++)
  {
    if (!isdigit ((int) token[i]))
      all_num = 0;
    if (token[i] >= 127 || iscntrl ((int) token[i])) {
      token[i] = 'z';
      all_num = 0;
    }
  }

  len = i - 1;

  if (isdigit ((int) token[0]))
  {
    if (token[len - 1] != '%')
      all_num = 1;
  }

  if (!isalnum ((int) token[0]) && token[0] != '$' && token[0] != '#')
    all_num = 1;

  if (is_received)
    all_num = 0;

  /* Ignore tokens that are all numbers, or contain high ASCII characters */
  if (all_num)
    return EINVAL;

  /* This is where we used to ignore certain headings */

  if (heading[0] != 0)
    snprintf (combined_token, sizeof (combined_token),
              "%s*%s", heading, token);
  else
    strlcpy (combined_token, token, sizeof (combined_token));

  crc = _ds_getcrc64 (combined_token);
#ifdef VERBOSE
  LOGDEBUG ("Token Hit: '%s'", combined_token);
#endif
  lht_hit (freq, crc, combined_token);

  if ((CTX->flags & DSF_CHAINED) && previous_token != NULL && !is_received)
  {
    snprintf (combined_token, sizeof (combined_token),
              "%s*%s+%s", heading, previous_token, token);
    crc = _ds_getcrc64 (combined_token);

    lht_hit (freq, crc, combined_token);
  }

  return 0;
}

int
_ds_process_body_token (DSPAM_CTX * CTX, char *token,
                        const char *previous_token, struct lht *freq)
{
  int all_num = 1, i;
  char combined_token[256];
  int len;
  unsigned long long crc;

  for (i = 0; token[i] != 0; i++)
  {
    if (!isdigit ((int) token[i]))
      all_num = 0;
    if (token[i] >= 127 || iscntrl ((int) token[i])) {
      token[i] = 'z';
      all_num = 0;
    }
  }

  len = i - 1;

  if (isdigit ((int) token[0]))
  {
    int l = len - 1;
    if (token[l] != '%')
      all_num = 1;
  }

  if (!isalnum ((int) token[0]) && token[0] != '$' && token[0] != '#')
    all_num = 1;

  /* Ignore tokens that are all numbers, or contain high ASCII characters */
  if (all_num)
    return EINVAL;

  crc = _ds_getcrc64 (token);

  lht_hit (freq, crc, token);

  if ((CTX->flags & DSF_CHAINED) && previous_token != NULL)
  {
    snprintf (combined_token, sizeof (combined_token), "%s+%s",
              previous_token, token);
    crc = _ds_getcrc64 (combined_token);

    lht_hit (freq, crc, combined_token);
  }

  return 0;
}

int
_ds_push_boundary (struct nt *stack, const char *boundary)
{
  char *y = malloc (strlen (boundary) + 3);
  if (y == NULL)
    return EUNKNOWN;

  sprintf (y, "--%s", boundary);
  nt_add (stack, (char *) y);
  return 0;
}

char *
_ds_pop_boundary (struct nt *stack)
{
  struct nt_node *node, *last_node = NULL, *parent_node = NULL;
  struct nt_c c;
  char *boundary = NULL;

  node = c_nt_first (stack, &c);
  while (node != NULL)
  {
    parent_node = last_node;
    last_node = node;
    node = c_nt_next (stack, &c);
  }
  if (parent_node != NULL)
    parent_node->next = NULL;
  else
    stack->first = NULL;

  if (last_node == NULL)
    return NULL;

  boundary = strdup (last_node->ptr);

  free (last_node->ptr);
  free (last_node);

  return boundary;
}

int
_ds_match_boundary (struct nt *stack, const char *buff)
{
  struct nt_node *node;
  struct nt_c c;

  node = c_nt_first (stack, &c);
  while (node != NULL)
  {
    if (!strncmp (buff, node->ptr, strlen (node->ptr)))
    {
      return 1;
    }
    node = c_nt_next (stack, &c);
  }
  return 0;
}

unsigned long long
_ds_getcrc64 (const char *s)
{
  static unsigned long long CRCTable[256];
  unsigned long long crc = 0;
  static int init = 0;

  if (!init)
  {
    int i;
    init = 1;
    for (i = 0; i <= 255; i++)
    {
      int j;
      unsigned long long part = i;
      for (j = 0; j < 8; j++)
      {
        if (part & 1)
          part = (part >> 1) ^ POLY64REV;
        else
          part >>= 1;
      }
      CRCTable[i] = part;
    }
  }
  for (; *s; s++)
  {
    unsigned long long temp1 = crc >> 8;
    unsigned long long temp2 =
      CRCTable[(crc ^ (unsigned long long) *s) & 0xff];
    crc = temp1 ^ temp2;
  }

  return crc;
}

/* Compute the complexity of a token */
int _ds_compute_complexity(const char *token) {
  int i, complexity = 1;
                                                                                
  if (token == NULL)
    return 1;
                                                                                
  for(i=0;token[i];i++) {
    if (token[i] == '+')
      complexity++;
  }
                                                                                
  return complexity;
}
